from collections.abc import AsyncGenerator
from typing import Any, Optional, TypedDict

from async_asgi_testclient import TestClient
from async_asgi_testclient.response import Response


class PostRequest(TypedDict, total=False):
    '''GraphQL POST Request
    https://graphql.org/learn/serving-over-http/#post-request
    '''

    query: str
    variables: dict[str, Any]
    operationName: str


class SubscribePayload(TypedDict):
    variables: dict[str, Any]
    extensions: dict[str, Any]
    operationName: Any
    query: str


class SubscribeMessage(TypedDict):
    '''GraphQL over WebSocket Protocol

    The type of the payload might depend on the value of the type.
    SubscribePayload might be only for the type "start".
    https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#subscribe
    https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md
    '''

    id: str
    type: str
    payload: SubscribePayload


async def gql_request(
    client: TestClient,
    query: str,
    variables: Optional[dict[str, Any]] = None,
) -> Any:
    resp = await gql_request_response(client, query, variables)
    result = resp.json()
    if err := result.get('errors'):
        raise Exception(err)
    return result['data']


async def gql_request_response(
    client: TestClient,
    query: str,
    variables: Optional[dict[str, Any]] = None,
) -> Response:
    request = PostRequest(query=query)
    if variables:
        request['variables'] = variables

    headers = {'Content-Type:': 'application/json'}

    return await client.post('/', json=request, headers=headers)


async def gql_subscribe(
    client: TestClient,
    query: str,
    variables: Optional[dict[str, Any]] = None,
) -> AsyncGenerator[Any, None]:
    payload = SubscribePayload(
        variables=variables or {},
        extensions={},
        operationName=None,
        query=query,
    )

    message = SubscribeMessage(id='1', type='start', payload=payload)

    async with client.websocket_connect('/') as ws:
        await ws.send_json(message)
        while True:
            resp_json = await ws.receive_json()
            if resp_json['type'] == 'complete':
                break
            yield resp_json['payload']['data']



from collections.abc import AsyncIterator, Sequence

from strawberry.exceptions import StrawberryGraphQLError
from strawberry.extensions import SchemaExtension
from strawberry.types.graphql import OperationType


class CORSMutation(SchemaExtension):
    def __init__(self, allow_origins: Sequence[str]):
        self._allow_origins = allow_origins

    async def on_execute(self) -> AsyncIterator[None]:
        if not self._is_origin_allowed():
            raise StrawberryGraphQLError('GraphQL Mutations not allowed')

        yield

    def _is_origin_allowed(self) -> bool:
        if '*' in self._allow_origins:
            return True

        op_type = self.execution_context.operation_type
        if op_type is not OperationType.MUTATION:
            return True

        request = self.execution_context.context.get('request')
        origin = request.headers.get('origin')
        if origin is None:
            #  Same origin or non-browser request
            return True

        if origin in self._allow_origins:
            return True

        return False



import asyncio
from collections.abc import AsyncIterator

import strawberry
from strawberry.types import Info

from nextline import Nextline
from nextlinegraphql.plugins.ctrl.cache import CacheStdout


@strawberry.type
class PromptingData:
    prompting: int
    file_name: str
    line_no: int
    trace_event: str


async def subscribe_counter() -> AsyncIterator[int]:
    for i in range(5):
        await asyncio.sleep(0)
        yield i + 1


def subscribe_state(info: Info) -> AsyncIterator[str]:
    nextline: Nextline = info.context["nextline"]
    return nextline.subscribe_state()


def subscribe_run_no(info: Info) -> AsyncIterator[int]:
    nextline: Nextline = info.context["nextline"]
    return nextline.subscribe_run_no()


def subscribe_trace_ids(info: Info) -> AsyncIterator[tuple[int, ...]]:
    nextline: Nextline = info.context["nextline"]
    return nextline.subscribe_trace_ids()


async def subscribe_prompting(
    info: Info, trace_id: int
) -> AsyncIterator[PromptingData]:
    nextline: Nextline = info.context["nextline"]
    async for i in nextline.subscribe_prompt_info_for(trace_id):
        if not i.file_name:
            # the initial yield at the beginning of a thread or task
            continue
        assert i.line_no is not None
        assert i.event is not None
        y = PromptingData(
            prompting=i.prompt_no if i.open else 0,
            file_name=i.file_name,
            line_no=i.line_no,
            trace_event=i.event,
        )
        yield y


def subscribe_stdout(info: Info) -> AsyncIterator[str]:
    cache_stdout = info.context['ctrl']['cache_stdout']
    assert isinstance(cache_stdout, CacheStdout)
    return cache_stdout.subscribe()


def subscribe_continuous_enabled(info: Info) -> AsyncIterator[bool]:
    nextline: Nextline = info.context["nextline"]
    return nextline.subscribe_continuous_enabled()


@strawberry.type
class Subscription:
    ctrl_counter: AsyncIterator[int] = strawberry.field(
        is_subscription=True, resolver=subscribe_counter
    )
    ctrl_state: AsyncIterator[str] = strawberry.field(
        is_subscription=True, resolver=subscribe_state
    )
    ctrl_run_no: AsyncIterator[int] = strawberry.field(
        is_subscription=True, resolver=subscribe_run_no
    )
    ctrl_trace_ids: AsyncIterator[tuple[int, ...]] = strawberry.field(
        is_subscription=True, resolver=subscribe_trace_ids
    )
    ctrl_prompting: AsyncIterator[PromptingData] = strawberry.field(
        is_subscription=True, resolver=subscribe_prompting
    )
    ctrl_stdout: AsyncIterator[str] = strawberry.field(
        is_subscription=True, resolver=subscribe_stdout
    )
    ctrl_continuous_enabled: AsyncIterator[bool] = strawberry.field(
        is_subscription=True, resolver=subscribe_continuous_enabled
    )



__all__ = ['spec', 'load_plugins']

from apluggy import PluginManager

from nextlinegraphql.plugins import ctrl, dev, graphql

from . import spec


def load_plugins(
    external: bool = True, plugins: list[object] | None = None
) -> PluginManager:
    '''Return a pluggy PluginManager with the plugins registered'''

    pm = PluginManager(spec.PROJECT_NAME)
    pm.add_hookspecs(spec)

    # The hooks are called in the reverse order of the plugin registration.
    # https://pluggy.readthedocs.io/en/stable/#call-time-order
    pm.register(graphql.Plugin(), name='graphql')
    pm.register(ctrl.Plugin(), name='ctrl')
    pm.register(dev.Plugin(), name='dev')
    # pm.set_blocked('schedule')

    if external:
        pm.load_setuptools_entrypoints(spec.PROJECT_NAME)

    if plugins:
        for plugin in plugins:
            pm.register(plugin)

    return pm



from typing import TYPE_CHECKING, Optional

import strawberry
from strawberry.types import Info

if TYPE_CHECKING:
    from nextline import Nextline


def query_hello(info: Info) -> str:
    request = info.context["request"]
    user_agent = request.headers.get("user-agent", "guest")
    return "Hello, %s!" % user_agent


def query_state(info: Info) -> str:
    nextline: Nextline = info.context["nextline"]
    return nextline.state or ''


def query_run_no(info: Info) -> int:
    nextline: Nextline = info.context["nextline"]
    return nextline.run_no


def query_trace_ids(info: Info) -> tuple[int, ...]:
    nextline: Nextline = info.context["nextline"]
    return nextline.trace_ids


def query_source(info: Info, file_name: Optional[str] = None) -> list[str]:
    nextline: Nextline = info.context["nextline"]
    return nextline.get_source(file_name)


def query_source_line(info: Info, line_no: int, file_name: Optional[str]) -> str:
    nextline: Nextline = info.context["nextline"]
    return nextline.get_source_line(line_no, file_name)


def query_exception(info: Info) -> Optional[str]:
    nextline: Nextline = info.context["nextline"]
    return nextline.format_exception()


def query_continuous_enabled(info: Info) -> bool:
    nextline: Nextline = info.context["nextline"]
    return nextline.continuous_enabled


@strawberry.type
class QueryCtrl:
    hello: str = strawberry.field(resolver=query_hello)
    state: str = strawberry.field(resolver=query_state)
    run_no: int = strawberry.field(resolver=query_run_no)
    trace_ids: tuple[int, ...] = strawberry.field(resolver=query_trace_ids)
    source: list[str] = strawberry.field(resolver=query_source)
    source_line: str = strawberry.field(resolver=query_source_line)
    exception: Optional[str] = strawberry.field(resolver=query_exception)
    continuous_enabled: bool = strawberry.field(resolver=query_continuous_enabled)


@strawberry.type
class Query:
    @strawberry.field
    def ctrl(self) -> QueryCtrl:
        return QueryCtrl()



from pathlib import Path
from typing import Optional

import strawberry
from strawberry.types import Info

from nextline import Nextline
from nextlinegraphql.plugins.ctrl import example_script as example_script_module

EXAMPLE_SCRIPT_PATH = Path(example_script_module.__file__).parent / 'script.py'
assert EXAMPLE_SCRIPT_PATH.is_file()
example_script = EXAMPLE_SCRIPT_PATH.read_text()


async def mutate_exec(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.run()
    return True


async def mutate_run_and_continue(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.run_and_continue()
    return True


async def mutate_reset(info: Info, statement: Optional[str] = None) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.reset(statement=statement)
    return True


async def mutate_send_pdb_command(
    info: Info, command: str, prompt_no: int, trace_no: int
) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.send_pdb_command(command, prompt_no, trace_no)
    return True


async def mutate_interrupt(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.interrupt()
    return True


async def mutate_terminate(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.terminate()
    return True


async def mutate_kill(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.kill()
    return True


async def mutate_load_example_script(info: Info) -> bool:
    nextline = info.context['nextline']
    assert isinstance(nextline, Nextline)
    await nextline.reset(statement=example_script)
    return True


@strawberry.type
class MutationCtrl:
    exec: bool = strawberry.field(resolver=mutate_exec)
    run_and_continue: bool = strawberry.field(resolver=mutate_run_and_continue)
    reset: bool = strawberry.field(resolver=mutate_reset)
    send_pdb_command: bool = strawberry.field(resolver=mutate_send_pdb_command)
    interrupt: bool = strawberry.field(resolver=mutate_interrupt)
    terminate: bool = strawberry.field(resolver=mutate_terminate)
    kill: bool = strawberry.field(resolver=mutate_kill)
    load_example_script: bool = strawberry.field(resolver=mutate_load_example_script)


@strawberry.type
class Mutation:
    @strawberry.field
    def ctrl(self) -> MutationCtrl:
        return MutationCtrl()



import asyncio
from threading import Thread


def in_thread():
    print("in_thread()")
    asyncio.run(in_task())


async def in_task():
    print("in_task()")


thread = Thread(target=in_thread)
thread.start()
thread.join()

print("here")



from collections.abc import AsyncIterator, Iterable, MutableMapping, Sequence
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, TypeAlias

import strawberry
from apluggy import PluginManager
from dynaconf import Dynaconf, Validator
from starlette.applications import Starlette
from starlette.types import ASGIApp
from strawberry.extensions import SchemaExtension
from strawberry.schema import BaseSchema
from strawberry.tools import merge_types

from nextlinegraphql.custom.strawberry import GraphQL
from nextlinegraphql.hook import spec

from .cors_mutation import CORSMutation
from .schema import Query

if TYPE_CHECKING:
    from strawberry.asgi import Request, Response, WebSocket

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (
    # NOTE: Provide default values `['*']` here so that these values will be overridden
    # by the settings file. If the default values were provided in `default.toml`, the
    # lists would be merged instead of overridden.
    Validator('GRAPHQL.MUTATION_ALLOW_ORIGINS', is_type_of=list, default=['*']),
)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[tuple[str, ...]]:
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf, hook: PluginManager) -> None:
        self._settings = settings
        self._app = _create_app(hook=hook, settings=settings)

    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, None, None)

    @spec.hookimpl(tryfirst=True)  # tryfirst so to be the outermost context
    @asynccontextmanager
    async def lifespan(self, app: Starlette) -> AsyncIterator[None]:
        app.mount('/', self._app)
        yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping[str, Any]) -> None:
        context['settings'] = self._settings


def _create_app(hook: PluginManager, settings: Dynaconf) -> ASGIApp:
    schema = _compose_schema(hook=hook, settings=settings)
    app = _EGraphQL(schema).set_hook(hook)
    return app


TRIO_LIST: TypeAlias = Iterable[tuple[type | None, type | None, type | None]]


def _compose_schema(hook: PluginManager, settings: Dynaconf) -> BaseSchema:
    # [(Query, Mutation, Subscription), ...]
    trio_list: TRIO_LIST = hook.hook.schema()

    Query, Mutation, Subscription = _merge_trios(trio_list)

    extensions = _create_scheme_extensions(settings=settings)

    schema = strawberry.Schema(
        query=Query, mutation=Mutation, subscription=Subscription, extensions=extensions
    )

    return schema


def _merge_trios(trio_list: TRIO_LIST) -> tuple[type, type | None, type | None]:
    # [(Query, ...), (Mutation, ...), (Subscription, ...)]
    transposed = list(map(tuple, zip(*trio_list)))
    transposed = [tuple(t for t in l if t) for l in transposed]  # remove None

    q, m, s = transposed
    assert q  # Query is required
    Query = merge_types('Query', q)
    Mutation = m and merge_types('Mutation', m) or None
    Subscription = s and merge_types('Subscription', s) or None

    return Query, Mutation, Subscription


def _create_scheme_extensions(settings: Dynaconf) -> Sequence[SchemaExtension]:
    mutation_allow_origins = settings.graphql.mutation_allow_origins
    extensions = [CORSMutation(allow_origins=mutation_allow_origins)]
    return extensions


class _EGraphQL(GraphQL):
    '''Extend the strawberry GraphQL app to override the `get_context` method

    This class is implemented in the way described in the strawberry document:
    https://strawberry.rocks/docs/integrations/asgi
    '''

    def set_hook(self, hook: PluginManager) -> '_EGraphQL':
        self._hook = hook
        return self

    async def get_context(
        self, request: 'Request | WebSocket', response: 'Response | WebSocket'
    ) -> Optional[Any]:
        context = {'request': request, 'response': response}
        self._hook.hook.update_strawberry_context(context=context)
        return context



import contextlib
import logging.config
from collections.abc import AsyncIterator
from logging import getLogger
from typing import Any

from apluggy import PluginManager
from dynaconf import Dynaconf
from rich import print
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

from .config import load_settings
from .hook import load_plugins


class App(Starlette):
    '''A wrap of ASGI App to attach hook and config.'''

    def set_hook(self, hook: PluginManager) -> 'App':
        self.hook = hook
        return self

    def set_config(self, config: Dynaconf) -> 'App':
        self.config = config
        return self


def create_app_for_test(
    extra_plugins: list[object] | None = None,
    enable_external_plugins: bool = False,
    enable_logging_configuration: bool = False,
    print_settings: bool = False,
) -> App:
    '''`create_app()` with default parameters suitable for tests.'''
    return create_app(
        extra_plugins=extra_plugins,
        enable_external_plugins=enable_external_plugins,
        enable_logging_configuration=enable_logging_configuration,
        print_settings=print_settings,
    )


def create_app(
    extra_plugins: list[object] | None = None,
    enable_external_plugins: bool = True,
    enable_logging_configuration: bool = True,
    print_settings: bool = True,
) -> App:
    '''App factory for Uvicorn.

    Parameters
    ----------
    extra_plugins
        Additional plugins to load. (Used for tests.)
    enable_external_plugins
        Do not load external plugins if False. (Used for tests.)
    enable_logging_configuration
        Leave logging configuration intact if False. (Used for tests.)
    print_settings
        Do not print settings if False. (Used for tests.)

    Returns
    -------
    Starlette
        The configured ASGI application instance.

    Notes
    -----
    Use the factory option to run with Uvicorn:

    $ uvicorn --factory nextlinegraphql:create_app

    See Also
    --------
    Uvicorn Application Factories : https://www.uvicorn.org/#application-factories

    '''

    hook, config = create_hook_and_config(
        extra_plugins=extra_plugins,
        enable_external_plugins=enable_external_plugins,
        enable_logging_configuration=enable_logging_configuration,
        print_settings=print_settings,
    )

    app = create_app_from(hook, config)
    return app


def create_hook_and_config(
    extra_plugins: list[object] | None,
    enable_external_plugins: bool,
    enable_logging_configuration: bool,
    print_settings: bool,
) -> tuple[PluginManager, Dynaconf]:
    hook = load_plugins(external=enable_external_plugins, plugins=extra_plugins)
    config = load_settings(hook)
    if print_settings:
        print('Settings:', config.as_dict())

    if enable_logging_configuration:
        configure_logging(config.logging)
    logger = getLogger(__name__)
    logger.info('Logging configured')

    hook.hook.configure(settings=config, hook=hook)

    plugin_names = [n for n, p in hook.list_name_plugin() if p]
    logger.info(f'Pluggy project name: {hook.project_name!r}')
    logger.info(f'Loaded plugins: {plugin_names}')
    return hook, config


def configure_logging(config: dict) -> None:
    logging.config.dictConfig(config)

    # https://pypi.org/project/logging_tree/
    # import logging_tree
    # logging_tree.printout()


def create_app_from(hook: PluginManager, config: Dynaconf) -> App:
    @contextlib.asynccontextmanager
    async def lifespan(app: App) -> AsyncIterator[None]:
        context = dict[Any, Any]()
        await hook.ahook.update_lifespan_context(app=app, hook=hook, context=context)
        async with hook.awith.lifespan(
            app=app, hook=hook, context=context
        ):  # pragma: no branch
            yield

    middleware = [
        Middleware(
            CORSMiddleware,
            allow_origins=config.cors['allow_origins'],
            allow_methods=['GET', 'POST', 'OPTIONS'],
            allow_headers=config.cors['allow_headers'],
            allow_credentials=config.cors['allow_credentials'],
        )
    ]

    app = (
        App(debug=True, lifespan=lifespan, middleware=middleware)
        .set_hook(hook)
        .set_config(config)
    )
    return app



import json

import strawberry
from dynaconf import Dynaconf
from strawberry.types import Info


def query_settings(info: Info) -> str:
    settings: Dynaconf = info.context['settings']
    return json.dumps(settings.to_dict())


@strawberry.type
class Query:
    settings: str = strawberry.field(resolver=query_settings)



__all__ = ['Plugin']

from collections.abc import AsyncIterator, MutableMapping
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Optional

from dynaconf import Dynaconf, Validator

from nextline import Nextline
from nextlinegraphql.hook import spec

from .cache import CacheStdout
from .example_script import statement
from .schema import Mutation, Query, Subscription

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (
    Validator('CTRL.TRACE_MODULES', must_exist=True, is_type_of=bool),
    Validator('CTRL.TRACE_THREADS', must_exist=True, is_type_of=bool),
)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[tuple[str, ...]]:
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf) -> None:
        self._trace_modules = settings.ctrl.trace_modules
        self._trace_threads = settings.ctrl.trace_threads

    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    async def update_lifespan_context(self, context: MutableMapping) -> None:
        self._nextline = Nextline(
            statement,
            trace_modules=self._trace_modules,
            trace_threads=self._trace_threads,
        )
        context['nextline'] = self._nextline

    @spec.hookimpl(trylast=True)  # trylast so to be the innermost context
    @asynccontextmanager
    async def lifespan(self) -> AsyncIterator[None]:
        '''Yield within the nextline context.'''
        self._cache_stdout = CacheStdout()
        self._nextline.register(self._cache_stdout)
        async with self._cache_stdout, self._nextline:
            yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['nextline'] = self._nextline
        ctrl = {'cache_stdout': self._cache_stdout}
        context['ctrl'] = ctrl
